/*
 NGR Converter - National Grid Reference to Latitude and Longitude
 Copyright © 2005-2015 Harry Whitfield

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin St, Fifth Floor, Boston, MA	 02110-1301	 USA

 NGR Converter - version 1.8.1
 1 May, 2015
 Copyright © 2005-2015 Harry Whitfield
 mailto:g6auc@arrl.net

 Based on an Ordnance Survey Information Booklet, March 1983.
*/

/*jslint browser, multivar, this */

/*property
    PI, abs, cos, eastings, latitude, longitude, northings, sin, sqrt, tan
*/


// Airy 1830: a = 6377563.396, b = 6356256.910

var n = 0.001673220250;
var e2 = 0.006670539762;
var a = 6375020.481; // 6377563.396 x F0
var b = 6353722.490; // 6356256.910 x F0
var F0 = 0.9996012717;

var Eorg = 400000.0;
var Norg = -100000.0;

var pi = Math.PI;

var n2 = n * n;
var n3 = n * n * n;

var Korg = 49.0 * pi / 180.0;
var Lorg = -2.0 * pi / 180.0;

function GridRef(eastings, northings) {
	"use strict";
    this.eastings = eastings;
    this.northings = northings;
}

function Position(latitude, longitude) {
	"use strict";
    this.latitude = latitude;
    this.longitude = longitude;
}

function fnm(k3, k4) {
	"use strict";
    var j3 = (1 + n + 1.25 * n2 + 1.25 * n3) * k3,
    	j4 = (3 * n + 3 * n2 + 2.625 * n3) * Math.sin(k3) * Math.cos(k4),
    	j5 = (1.875 * n2 + 1.875 * n3) * Math.sin(2 * k3) * Math.cos(2 * k4),
    	j6 = (35.0 / 24.0) * n3 * Math.sin(3 * k3) * Math.cos(3 * k4);

    return (b * (j3 - j4 + j5 - j6));
}

function fnv(K) {
	"use strict";
    var s = Math.sin(K);

    return (a / Math.sqrt(1 - e2 * s * s));
}

function fnr(V, K) {
	"use strict";
    var s = Math.sin(K);

    return (V * (1 - e2) / (1 - e2 * s * s));
}

function ngr(K, L) { // lat,long -> eastings,northings
	"use strict";
    var M = fnm(K - Korg, K + Korg),
    	V = fnv(K),
    	R = fnr(V, K),
    	H2 = V / R - 1.0,
    	s = Math.sin(K),
    	c = Math.cos(K),
    	t = Math.tan(K),
    	t2 = t * t,
    	t4 = t2 * t2,
    	c2 = c * c,
    	c3 = c * c2,
    	c5 = c2 * c3,
    	P = L - Lorg,
    	j3 = M + Norg,
    	j4 = 0.5 * V * s * c,
    	j5 = (V / 24) * s * c3 * (5 - t2 + 9 * H2),
    	j6 = (V / 720) * s * c5 * (61 - 58 * t2 + t4),
    	N = j3 + P * P * j4 + P * P * P * P * j5 + P * P * P * P * P * P * j6,
    	j7 = V * c,
    	j8 = (V / 6) * c3 * (V / R - t2),
    	j9 = (V / 120) * c5 * (5 - 18 * t2 + t4 + 14 * H2 - 58 * t2 * H2),
    	E = Eorg + P * j7 + P * P * P * j8 + P * P * P * P * P * j9;
    return new GridRef(E, N);
}

function phi(N) {
	"use strict";
    var K = Korg,
    	M = 0;

    do {
        K = K + (N - Norg - M) / a;
        M = fnm(K - Korg, K + Korg);
    } while (Math.abs(N - Norg - M) >= 0.001);
    return K;
}

function lat_long(E, N) { // ngr -> latitude,longitude
	"use strict";
    var K = phi(N),
    	c = Math.cos(K),
    	t = Math.tan(K),
    	t2 = t * t,
    	t4 = t2 * t2,
    	t6 = t2 * t4,
    	V = fnv(K),
    	R = fnr(V, K),
    	H2 = V / R - 1.0,
    	y = E - Eorg,
    	sV = 0.000001 * V,
		sR = 0.000001 * R,
    	sy = 0.000001 * y,
    	j3 = t / (2 * sR * sV),
    	j4 = t / (24 * sR * sV * sV * sV) * (5 + 3 * t2 + H2 - 9 * t2 * H2),
    	j5 = t / (720 * sR * sV * sV * sV * sV * sV) * (61 + 90 * t2 + 45 * t4),
    	k9 = K - sy * sy * j3 + sy * sy * sy * sy * j4 - sy * sy * sy * sy * sy * sy * j5,
    	j6 = 1.0 / (c * sV),
    	j7 = 1.0 / (c * 6 * sV * sV * sV) * (sV / sR + 2 * t2),
    	j8 = 1.0 / (c * 120 * sV * sV * sV * sV * sV) * (5 + 28 * t2 + 24 * t4),
    	j9 = 1.0 / (c * 5040 * sV * sV * sV * sV * sV * sV * sV),
    	sj9 = j9 * (61 + 662 * t2 + 1320 * t4 + 720 * t6),
    	LL = Lorg + sy * j6 - sy * sy * sy * j7 + sy * sy * sy * sy * sy * j8 - sy * sy * sy * sy * sy * sy * sy * sj9,
    	KK = k9;

    return new Position(KK, LL);
}

function fnc(K, L) {
	"use strict";
    var V = fnv(K),
    	R = fnr(V, K),
    	H2 = V / R - 1.0,
    	s = Math.sin(K),
    	c = Math.cos(K),
    	t = Math.tan(K),
    	t2 = t * t,
    	c2 = c * c,
    	c4 = c2 * c2,
    	P = L - Lorg,
    	j4 = s * c2 / 3.0 * (1 + 3 * H2 + 2 * H2 * H2),
    	j5 = s * c4 / 15.0 * (2 - t2);

    return (P * s + P * P * P * j4 + P * P * P * P * P * j5);
}

function con(E, N) {
	"use strict";
    var K = phi(N),
    	t = Math.tan(K),
    	t2 = t * t,
    	t4 = t2 * t2,
    	V = fnv(K),
    	R = fnr(V, K),
    	H2 = V / R - 1.0,
    	y = E - Eorg,
    	sV = 0.000001 * V,
//    	sR = 0.000001 * R,
    	sy = 0.000001 * y,
    	j3 = t / sV,
    	j4 = t / (3 * sV * sV * sV) * (1 + t2 - H2 - 2 * H2 * H2),
    	j5 = t / (15 * sV * sV * sV * sV * sV) * (2 + 5 * t2 + 3 * t4);

    return (sy * j3 - sy * sy * sy * j4 + sy * sy * sy * sy * sy * j5);
}

function fnf(K, L) {
	"use strict";
    var V = fnv(K),
    	R = fnr(V, K),
    	H2 = V / R - 1.0,
    	c = Math.cos(K),
    	t = Math.tan(K),
    	t2 = t * t,
    	c2 = c * c,
    	c4 = c2 * c2,
    	P = L - Lorg,
    	j3 = c2 / 2.0 * (1 + H2),
    	j4 = c4 / 24.0 * (5 - 4 * t2 + 14 * H2 - 28 * t2 * H2);

    return (F0 * (1 + P * P * j3 + P * P * P * P * j4));
}

function fac(E, N) {
	"use strict";
    var K = phi(N),
    	V = fnv(K),
    	R = fnr(V, K),
    	H2 = V / R - 1.0,
    	y = E - Eorg,
    	sV = 0.000001 * V,
    	sR = 0.000001 * R,
    	sy = 0.000001 * y,
    	j3 = 1.0 / (2 * sR * sV),
    	j4 = (1 + 4 * H2) / (24 * sR * sR * sV * sV);

    return (F0 * (1 + sy * sy * j3 + sy * sy * sy * sy * j4));
}

function deltaTa(Ea, Na, Eb, Nb) {
	"use strict";
    var K = phi(0.5 * (Na + Nb)),
    	V = fnv(K),
    	R = fnr(V, K),
//    	H2 = V / R - 1.0,
    	ya = Ea - Eorg,
    	yb = Eb - Eorg,
    	sV = 0.001 * V,
    	sR = 0.001 * R,
    	sya = 0.000001 * ya,
    	syb = 0.000001 * yb,
    	j3 = 1.0 / (6 * sR * sV);

    return ((2 * sya + syb) * (Na - Nb) * j3);
}

function deltaTb(Ea, Na, Eb, Nb) {
	"use strict";
    var K = phi(0.5 * (Na + Nb)),
    	V = fnv(K),
    	R = fnr(V, K),
//    	H2 = V / R - 1.0,
    	ya = Ea - Eorg,
    	yb = Eb - Eorg,
    	sV = 0.001 * V,
    	sR = 0.001 * R,
    	sya = 0.000001 * ya,
    	syb = 0.000001 * yb,
    	j3 = 1.0 / (6 * sR * sV);

    return ((sya + 2 * syb) * (Nb - Na) * j3);
}
